package fr.mrdev023.vulkan_java.vk;

import org.lwjgl.PointerBuffer;
import org.lwjgl.system.MemoryStack;
import org.lwjgl.vulkan.*;
import org.tinylog.Logger;

import fr.mrdev023.vulkan_java.vk.utils.SuitablePhysicalDeviceFinder;

import java.nio.*;

import static fr.mrdev023.vulkan_java.vk.VulkanError.vkCheck;
import static org.lwjgl.vulkan.VK11.*;

public class Device {

    private final PhysicalDevice physicalDevice;
    private final VkDevice vkDevice;
    private final Queue.PresentQueue presentQueue;
    private final Queue.GraphicsQueue graphicsQueue;
    private final Queue.ComputeQueue computeQueue;
    private final Queue.TransferQueue transferQueue;

    public Device(SuitablePhysicalDeviceFinder.MatchResult physicalDeviceMatch) throws VulkanError {
        this.physicalDevice = physicalDeviceMatch.physicalDevice;

        try (MemoryStack stack = MemoryStack.stackPush()) {
            var extensions = physicalDeviceMatch.validatorResult.deviceExtensionsResult.matchingExtensions;
            PointerBuffer requiredExtensions = stack.mallocPointer(extensions.size());
            for (String extension : extensions) {
                requiredExtensions.put(stack.ASCII(extension));
            }
            requiredExtensions.flip();

            // Set up required features
            VkPhysicalDeviceFeatures features = VkPhysicalDeviceFeatures.calloc(stack);

            // Enable all the queue families
            VkQueueFamilyProperties.Buffer queuePropsBuff = physicalDevice.getVkQueueFamilyProps();
            int numQueuesFamilies = queuePropsBuff.capacity();
            VkDeviceQueueCreateInfo.Buffer queueCreationInfoBuf = VkDeviceQueueCreateInfo.calloc(numQueuesFamilies,
                    stack);
            for (int i = 0; i < numQueuesFamilies; i++) {
                FloatBuffer priorities = stack.callocFloat(queuePropsBuff.get(i).queueCount());
                queueCreationInfoBuf.get(i)
                        .sType(VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO)
                        .queueFamilyIndex(i)
                        .pQueuePriorities(priorities);
            }

            VkDeviceCreateInfo deviceCreateInfo = VkDeviceCreateInfo.calloc(stack)
                    .sType(VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO)
                    .ppEnabledExtensionNames(requiredExtensions)
                    .pEnabledFeatures(features)
                    .pQueueCreateInfos(queueCreationInfoBuf);

            PointerBuffer pp = stack.mallocPointer(1);
            vkCheck(vkCreateDevice(physicalDevice.getVkPhysicalDevice(), deviceCreateInfo, null, pp),
                    "Failed to create device");
            vkDevice = new VkDevice(pp.get(0), physicalDevice.getVkPhysicalDevice(), deviceCreateInfo);

            // graphicsQueue = new Queue.GraphicsQueue(this,
            // physicalDeviceMatch.graphicsQueueFamilyIndex, 0);
            // computeQueue = new Queue.ComputeQueue(this,
            // physicalDeviceMatch.computeQueueFamilyIndex, 1);
            // transferQueue = new Queue.TransferQueue(this,
            // physicalDeviceMatch.transferQueueFamilyIndex, 2);
            presentQueue = null;
            graphicsQueue = null;
            computeQueue = null;
            transferQueue = null;
        }

        Logger.debug("Vulkan device created");
    }

    public void destroy() {
        Logger.debug("Destroying Vulkan device");
        vkDestroyDevice(vkDevice, null);
    }

    public PhysicalDevice getPhysicalDevice() {
        return physicalDevice;
    }

    public VkDevice getVkDevice() {
        return vkDevice;
    }

    public void waitIdle() {
        vkDeviceWaitIdle(vkDevice);
    }

    public Queue.PresentQueue getPresentQueue() {
        return presentQueue;
    }

    public Queue.GraphicsQueue getGraphicsQueue() {
        return graphicsQueue;
    }

    public Queue.ComputeQueue getComputeQueue() {
        return computeQueue;
    }

    public Queue.TransferQueue getTransferQueue() {
        return transferQueue;
    }
}